-- Novation Launchkey MK4 Reason Script -
-- v1.0.0.7

--------------------
-- user functions --
--------------------
     

-- this must be global cause it's used to directly read values from pads
function read_pad(x)
   -- turn pads into buttons - return 0-1
   
   if x ~= 0 then
      return 1
   end
   
   return x
end

local function stringToHex(s)
   s = s or ""
   local output = ""
   for i=1, #s do
      output = output .. string.format("%X ", string.byte(s, i, i))
   end
   
   return output
end

--[[
local function update_encoder(index, value)
   return remote.make_midi("BF " .. stringToHex(tostring(index)) .. " " .. stringToHex(tostring(value)))
end  
]]--

local function clear_display()
   return remote.make_midi("F0 00 20 29 02 " .. PN .. " 04 20 00 F7")
end



local function get_control_index(ctrl)
   -- used in set_param_name() and set_param_value()
   
   ctrl = ctrl or "pot 0"

   local control = {["pot 0"] = 21, ["pot 1"] = 22, ["pot 2"] = 23, ["pot 3"] = 24,
      ["pot 4"] = 25, ["pot 5"] = 26, ["pot 6"] = 27, ["pot 7"] = 28,
      ["fader 0"] = 5, ["fader 1"] = 6, ["fader 2"] = 7, ["fader 3"] = 8,
      ["fader 4"] = 9, ["fader 5"] = 10, ["fader 6"] = 11, ["fader 7"] = 12,
      ["fader 8"] = 13}
   -- don't care if it's nil
   return control[ctrl]
   
end




local function dec_to_hex(input)
   -- take an integer and turn it into a string of HEX digits
   -- ie 255 = "32 35 35"
   
   input = input or 0
   input = tostring(input)

   -- Lua 5.1 doesn't have table.pack
   local input_tbl = {}
   for i=1, #input do
      input_tbl[i] = string.byte(input, i, i)
   end
   
   local output = ""
   for k, v in ipairs(input_tbl) do
      output = output ..  string.format("%02X ", v)
   end

   return output
end


local function set_param_name(p_name, ctrl_id)
   -- args: p_name = parameter name (ascii), ctrl_id = control as string (ie "pot 3" or "fader 8") 
   
   -- if ctrl_id is invalid return "false"
   ctrl_id = get_control_index(ctrl_id)
   if ctrl_id == nil or type(ctrl_id) ~= "number" then return false end

   -- make sure there are always 2 digits
   -- even for numbers < 16 (0x00 - 0x0F)
   ctrl_id = string.format("%02X", ctrl_id)
   
   p_name = p_name or ""
   local asciitext = stringToHex(p_name)
   
   return remote.make_midi("F0 00 20 29 02 " .. PN .. " 04 " .. ctrl_id .. " 61 F7 F0 00 20 29 02 " .. PN .. " 06 " .. ctrl_id .. " 00 " .. asciitext .. " F7 ")
      --("F0 00 20 29 02 " .. PN .. " 07 " .. ctrl_id .. " " .. asciitext .. " F7")
end


local function set_param_value(p_value, ctrl_id)
   -- args: p_value = parameter value, ctrl_id = control as string (ie "pot 3" or "fader 8") 
   
   -- if ctrl_id is invalid return "false"
   ctrl_id = get_control_index(ctrl_id)
   if ctrl_id == nil or type(ctrl_id) ~= "number" then return false end

   -- make sure there are always 2 digits
   -- even for numbers < 16 (0x00 - 0x0F)
   ctrl_id = string.format("%02X", ctrl_id)
   
   p_value = p_value or 0
   p_value = dec_to_hex(p_value)
   
   return remote.make_midi("F0 00 20 29 02 " .. PN .. " 06 " .. ctrl_id .. " 01 " .. p_value .. " F7")
end


local function set_display(row, text)
   if row == 0 then
      row = "00"
   elseif row == 1 then
      row = "01"
   end

   if text ~= "" then
      text = stringToHex(text)
   end

   return remote.make_midi(
      "F0 00 20 29 02 " .. PN .. " 04 21 01 F7 " ..  -- set screen
      "F0 00 20 29 02 " .. PN .. " 06 21 " .. row .. " " .. text .. "  F7 " ..  -- text to display on specific row
      "F0 00 20 29 02 " .. PN .. " 04 21 7F F7"  -- triggeer the display 
      )
end




local function match_events(mask,events)

   for i, event in ipairs(events) do
      local res = remote.match_midi(mask,event)
      if res ~= nil then
	      return true
      end
   end
   return false
end





local function mute_control(data_value, curr_value, last_value)
   -- check whether the control has crossed the parameter value point
   -- used for pot/fader pickup
   
   if last_value ~= -1 then
      if last_value < curr_value then
	     if data_value >= curr_value then
            return true
         else
            return false
         end
      elseif last_value > curr_value then
            if data_value <= curr_value then
               return true
            else
   	        return false
            end
      elseif last_value == curr_value then
         return true
      end

   -- in case init status with 'last_value' == -1
   else
      if math.abs(data_value - curr_value) < 10 then
	     return true
      else
	     return false
      end
   end
end



------------------------------------
-- defining control surface items --
------------------------------------

function remote_init(manufacturer, model)

   g_model = "Launchkey MK4"
   local items={
      
      {name="Knob 1", input="value", output="value", min=0, max=127},
      {name="Knob 2", input="value", output="value", min=0, max=127},
      {name="Knob 3", input="value", output="value", min=0, max=127},
      {name="Knob 4", input="value", output="value", min=0, max=127},
      {name="Knob 5", input="value", output="value", min=0, max=127},
      {name="Knob 6", input="value", output="value", min=0, max=127},
      {name="Knob 7", input="value", output="value", min=0, max=127},
      {name="Knob 8", input="value", output="value", min=0, max=127},
      {name="STOP", input="button", output="value", min=0, max=127},
      {name="PLAY", input="button", output="value", min=0, max=127},
      {name="LOOP", input="button"},
      {name="RECORD", input="button", output="value", min=0, max=127},
      {name="Pad 1", input="button", output="value"},
      {name="Pad 2", input="button", output="value"},
      {name="Pad 3", input="button", output="value"},
      {name="Pad 4", input="button", output="value"},
      {name="Pad 5", input="button", output="value"},
      {name="Pad 6", input="button", output="value"},
      {name="Pad 7", input="button", output="value"},
      {name="Pad 8", input="button", output="value"},
      {name="Pad 9", input="button", output="value"},
      {name="Pad 10", input="button", output="value"},
      {name="Pad 11", input="button", output="value"},
      {name="Pad 12", input="button", output="value"},
      {name="Pad 13", input="button", output="value"},
      {name="Pad 14", input="button", output="value"},
      {name="Pad 15", input="button", output="value"},
      {name="Pad 16", input="button", output="value"},
      {name="Cliplaunch 1", input="button"},
      {name="Cliplaunch 2", input="button"},
      {name="Clip Up", input="button", output="value"},
      {name="Clip Down", input="button", output="value"}, 
      {name="Track Up", input="button"},    -- track right
      {name="Track Down", input="button"},  -- track left
      {name="Intelligent Knobs", input="button"},
      {name="Intelligent Pads", input="button"},
      {name="patch_name", output="text"},
      {name="group_name", output="text"},
      {name="device_name", output="text"},
      {name="track_name", output="text"},
      {name="Fader 1", input="value", output="value", min=0, max=127},
      {name="Fader 2", input="value", output="value", min=0, max=127},
      {name="Fader 3", input="value", output="value", min=0, max=127},
      {name="Fader 4", input="value", output="value", min=0, max=127},
      {name="Fader 5", input="value", output="value", min=0, max=127},
      {name="Fader 6", input="value", output="value", min=0, max=127},
      {name="Fader 7", input="value", output="value", min=0, max=127},
      {name="Fader 8", input="value", output="value", min=0, max=127},
      {name="Fader 9", input="value", output="value", min=0, max=127},
      {name="Keyboard", input="keyboard"},
      {name="Pitch Bend", input="value", min=0, max=16383},
      {name="Modulation", input="value", min=0, max=127},
      {name="Sustain", input="value", min=0, max=127},
      {name="Shift", input="value", min=0, max=127},
      {name="CLICK", input="button", output="value", min=0, max=127},
      {name="Quantize", input="button", output="value"},
      {name="Undo", input="button", output="value"},
      {name="LCD", output="text"},

   }
   remote.define_items(items)

   g_knob_1 = 1
   g_knob_2 = 2
   g_knob_3 = 3
   g_knob_4 = 4
   g_knob_5 = 5
   g_knob_6 = 6
   g_knob_7 = 7
   g_knob_8 = 8
   g_stop = 9
   g_play = 10
   g_record = 12
   g_pad_1 = 13
   g_pad_2 = 14
   g_pad_3 = 15
   g_pad_4 = 16
   g_pad_5 = 17
   g_pad_6 = 18
   g_pad_7 = 19
   g_pad_8 = 20
   g_pad_9 = 21
   g_pad_10 = 22
   g_pad_11 = 23
   g_pad_12 = 24
   g_pad_13 = 25
   g_pad_14 = 26
   g_pad_15 = 27
   g_pad_16 = 28
   g_cliplaunch_1 = 30
   g_cliplaunch_2 = 31 		
   g_clip_up = 32
   g_clip_down = 33	
   g_track_up = 34
   g_track_down = 35
   g_button_1 = 151
   g_button_2 = 152
   g_button_3 = 153
   g_button_4 = 154
   g_button_5 = 155
   g_button_6 = 156
   g_button_7 = 157
   g_button_8 = 158
   g_button_9 = 159
   g_fader_1 = 41
   g_fader_2 = 42
   g_fader_3 = 43
   g_fader_4 = 44
   g_fader_5 = 45
   g_fader_6 = 46
   g_fader_7 = 47
   g_fader_8 = 48
   g_fader_9 = 49
   g_lcd = table.getn(items)

   PN = "14" -- LK4 Standard


   

   local inputs={
      {pattern="BF 05 xx", name="Fader 1", port=1},
      {pattern="BF 06 xx", name="Fader 2", port=1},
      {pattern="BF 07 xx", name="Fader 3", port=1},
      {pattern="BF 08 xx", name="Fader 4", port=1},
      {pattern="BF 09 xx", name="Fader 5", port=1},
      {pattern="BF 0A xx", name="Fader 6", port=1},
      {pattern="BF 0B xx", name="Fader 7", port=1},
      {pattern="BF 0C xx", name="Fader 8", port=1},
      {pattern="BF 0D xx", name="Fader 9", port=1}, 
      {pattern="b0 15 xx", name="Knob 1", port=1},
      {pattern="b0 16 xx", name="Knob 2", port=1},
      {pattern="b0 17 xx", name="Knob 3", port=1},
      {pattern="b0 18 xx", name="Knob 4", port=1},
      {pattern="b0 19 xx", name="Knob 5", port=1},
      {pattern="b0 1A xx", name="Knob 6", port=1},
      {pattern="b0 1B xx", name="Knob 7", port=1},
      {pattern="b0 1C xx", name="Knob 8", port=1},
      {pattern="B0 74 xx", name="STOP", port=1},
      {pattern="B0 73 xx", name="PLAY", port=1},  
      {pattern="B0 76 xx", name="LOOP", port=1},
      {pattern="B0 75 xx", name="RECORD", port=1},  
      {pattern="90 60 xx", name="Pad 1", value="read_pad(x)", port=1},
      {pattern="90 61 xx", name="Pad 2", value="read_pad(x)", port=1},
      {pattern="90 62 xx", name="Pad 3", value="read_pad(x)", port=1},
      {pattern="90 63 xx", name="Pad 4", value="read_pad(x)", port=1},
      {pattern="90 64 xx", name="Pad 5", value="read_pad(x)", port=1},
      {pattern="90 65 xx", name="Pad 6", value="read_pad(x)", port=1},
      {pattern="90 66 xx", name="Pad 7", value="read_pad(x)", port=1},
      {pattern="90 67 xx", name="Pad 8", value="read_pad(x)", port=1},
      {pattern="90 70 xx", name="Pad 9", value="read_pad(x)", port=1},
      {pattern="90 71 xx", name="Pad 10", value="read_pad(x)", port=1},
      {pattern="90 72 xx", name="Pad 11", value="read_pad(x)", port=1},
      {pattern="90 73 xx", name="Pad 12", value="read_pad(x)", port=1},
      {pattern="90 74 xx", name="Pad 13", value="read_pad(x)", port=1},
      {pattern="90 75 xx", name="Pad 14", value="read_pad(x)", port=1},
      {pattern="90 76 xx", name="Pad 15", value="read_pad(x)", port=1},
      {pattern="90 77 xx", name="Pad 16", value="read_pad(x)", port=1},
      --[[
	 DRUM LAYOUT
	 {pattern="99 28 ?<???x>", name="Pad 1", port=1},
	 {pattern="99 29 ?<???x>", name="Pad 2", port=1},
	 {pattern="99 2A ?<???x>", name="Pad 3", port=1},
	 {pattern="99 2B ?<???x>", name="Pad 4", port=1},
	 {pattern="99 30 ?<???x>", name="Pad 5", port=1},
	 {pattern="99 31 ?<???x>", name="Pad 6", port=1},
	 {pattern="99 32 ?<???x>", name="Pad 7", port=1},
	 {pattern="99 33 ?<???x>", name="Pad 8", port=1},
	 {pattern="99 24 ?<???x>", name="Pad 9", port=1},
	 {pattern="99 25 ?<???x>", name="Pad 10", port=1},
	 {pattern="99 26 ?<???x>", name="Pad 11", port=1},
	 {pattern="99 27 ?<???x>", name="Pad 12", port=1},
	 {pattern="99 2C ?<???x>", name="Pad 13", port=1},
	 {pattern="99 2D ?<???x>", name="Pad 14", port=1},
	 {pattern="99 2E ?<???x>", name="Pad 15", port=1},
	 {pattern="99 2F ?<???x>", name="Pad 16", port=1},
      ]]--
      {pattern="B0 68 ?<???x>", name="Cliplaunch 1", port=1}, -- fake MIDI msg
      {pattern="B0 69 ?<???x>", name="Cliplaunch 2", port=1}, -- fake MIDI msg
      {pattern="B0 6B ?<???x>", name="Clip Up", port=1}, -- fake MIDI msg
      {pattern="B0 6A ?<???x>", name="Clip Down", port=1}, -- fake MIDI msg
      {pattern="B0 66 ?<???x>", name="Track Up", port=1},
      {pattern="B0 67 ?<???x>", name="Track Down", port=1},
      {pattern="bF 0d ?<???x>", name="Intelligent Knobs", port=1},
      {pattern="bF 0f ?<???x>", name="Intelligent Pads", port=1},
      {pattern="e? xx yy", name="Pitch Bend", value="y*128 + x", port=2},
      {pattern="b? 01 xx", name="Modulation", port=2},
      {pattern="<100x>? yy zz", name="Keyboard", port=2},
      {pattern="b? 40 xx", name="Sustain", port=2},
      {pattern="B6 3F xx", name="Shift", port=1},
      {pattern="B0 4C ?<???x>", name="CLICK", port=1},
      {pattern="BF 4B xx", name="Quantize", port=1},
      {pattern="BF 4D xx", name="Undo", port=1},
   }
   remote.define_auto_inputs(inputs)


   
   local outputs = {
      {name="Pad 1", pattern="90 60 0<00xx>", x="value*3"},
      {name="Pad 2", pattern="90 61 0<00xx>", x="value*3"},
      {name="Pad 3", pattern="90 62 0<00xx>", x="value*3"},
      {name="Pad 4", pattern="90 63 0<00xx>", x="value*3"},
      {name="Pad 5", pattern="90 64 0<00xx>", x="value*3"},
      {name="Pad 6", pattern="90 65 0<00xx>", x="value*3"},
      {name="Pad 7", pattern="90 66 0<00xx>", x="value*3"},
      {name="Pad 8", pattern="90 67 0<00xx>", x="value*3"},
      {name="Pad 9", pattern="90 70 0<00xx>", x="value*3"},
      {name="Pad 10", pattern="90 71 0<00xx>", x="value*3"},
      {name="Pad 11", pattern="90 72 0<00xx>", x="value*3"},
      {name="Pad 12", pattern="90 73 0<00xx>", x="value*3"},
      {name="Pad 13", pattern="90 74 0<00xx>", x="value*3"},
      {name="Pad 14", pattern="90 75 0<00xx>", x="value*3"},
      {name="Pad 15", pattern="90 76 0<00xx>", x="value*3"},
      {name="Pad 16", pattern="90 77 0<00xx>", x="value*3"},
      {name="Clip Up", pattern="BF 6A xx", x="value"},
      {name="Clip Down", pattern="BF 6B xx", x="value"},
      {name="Knob 1", pattern="BF 15 xx", x="value"},
      {name="Knob 2", pattern="BF 16 xx", x="value"},
      {name="Knob 3", pattern="BF 17 xx", x="value"},
      {name="Knob 4", pattern="BF 18 xx", x="value"},
      {name="Knob 5", pattern="BF 19 xx", x="value"},
      {name="Knob 6", pattern="BF 1A xx", x="value"},
      {name="Knob 7", pattern="BF 1B xx", x="value"},
      {name="Knob 8", pattern="BF 1C xx", x="value"},
--[[       {name="Fader 1", pattern="BF 05 xx", x="value"},
      {name="Fader 2", pattern="BF 06 xx", x="value"},
      {name="Fader 3", pattern="BF 07 xx", x="value"},
      {name="Fader 4", pattern="BF 08 xx", x="value"},
      {name="Fader 5", pattern="BF 09 xx", x="value"},
      {name="Fader 6", pattern="BF 0A xx", x="value"},
      {name="Fader 7", pattern="BF 0B xx", x="value"},
      {name="Fader 8", pattern="BF 0C xx", x="value"},
      {name="Fader 9", pattern="BF 0D xx", x="value"}, ]]
      {name="PLAY", pattern="B0 73 xx", x="value"},
      {name="RECORD", pattern="B0 75 xx", x="value"},
   }
   remote.define_auto_outputs(outputs)
end


-----------------------------
-- auto detect the surface --
-----------------------------

function remote_probe(manufacturer,model,prober)

   local results={}
   local port_out = 0
   local ins = {}
   local received_ports = {}
   local dev_found = 0
   local request_events = {}
   local response = ""

   request_events = { remote.make_midi("F0 7E 7F 06 01 F7") }
   response = {"F0 7E 00 06 02 00 20 29 43 01 00 01 ?? ?? ?? ?? F7",
               "F0 7E 00 06 02 00 20 29 44 01 00 01 ?? ?? ?? ?? F7",
               "F0 7E 00 06 02 00 20 29 45 01 00 01 ?? ?? ?? ?? F7",
               "F0 7E 00 06 02 00 20 29 46 01 00 01 ?? ?? ?? ?? F7"}
   
   -- check all the MIDI OUT ports
   for outPortIndex = 1, prober.out_ports do
      -- send device inquiry msg
      prober.midi_send_function(outPortIndex, request_events)
      prober.wait_function(50)


      -- check all the MIDI IN ports
      for inPortIndex = 1, prober.in_ports do
         local events = prober.midi_receive_function(inPortIndex)
         if match_events(response[1], events) or match_events(response[2], events) or match_events(response[3], events) or match_events(response[4], events) then
            port_out = outPortIndex         -- DAW port
            table.insert(ins, outPortIndex)  -- DAW port
            table.insert(ins, outPortIndex - 1)      -- MIDI port
            dev_found = dev_found + 1
            break
         end
      end
      if dev_found ~= 0 then
         break
      end
   end
      
   -- check a device has been found
   if dev_found ~= 0 then
      local one_result={ in_ports={ins[1], ins[2]}, out_ports={port_out} }
      table.insert(results, one_result)
   end
   return results
end


------------------------------
-- set device to 'DAW mode' --
-- and enter SESSION layout --
------------------------------

function remote_prepare_for_use()
   local retEvents = {
      remote.make_midi("9F 0C 7F"), -- ENTER DAW MODE
      remote.make_midi("B6 40 02"), -- PAD: SESSION LAYOUT
      remote.make_midi("B6 41 02"), -- KNOBS: DEVICE LAYOUT 
      --remote.make_midi("B6 42 02"), -- FADERS: VOLUME LAYOUT

      remote.make_midi("F0 00 20 29 02 " .. PN .. " 04 20 02 F7"), -- CONFIGURE STATIONARY DISPLAY 
      remote.make_midi("F0 00 20 29 02 " .. PN .. " 06 20 01 52 65 61 73 6F 6E F7"), -- SET TEXT "REASON"
      remote.make_midi("F0 00 20 29 02 " .. PN .. " 04 20 7F F7"), -- TRIGGER STATIONARY DISPLAY      
   }
   return retEvents
end


-------------------
-- exit DAW mode --
-------------------

function remote_release_from_use()

   local retEvents = {
      remote.make_midi("9F 0C 00"),
   }
   return retEvents
end


-- initialize params
g_fader_1_last_value = -1
g_fader_2_last_value = -1
g_fader_3_last_value = -1
g_fader_4_last_value = -1
g_fader_5_last_value = -1
g_fader_6_last_value = -1
g_fader_7_last_value = -1
g_fader_8_last_value = -1
g_fader_9_last_value = -1
g_knob_1_last_value = -1
g_knob_2_last_value = -1
g_knob_3_last_value = -1
g_knob_4_last_value = -1
g_knob_5_last_value = -1
g_knob_6_last_value = -1
g_knob_7_last_value = -1
g_knob_8_last_value = -1
g_fader_1_current_value = 0
g_fader_2_current_value = 0
g_fader_3_current_value = 0
g_fader_4_current_value = 0
g_fader_5_current_value = 0
g_fader_6_current_value = 0
g_fader_7_current_value = 0
g_fader_8_current_value = 0
g_fader_9_current_value = 0
g_knob_1_current_value = 0
g_knob_2_current_value = 0
g_knob_3_current_value = 0
g_knob_4_current_value = 0
g_knob_5_current_value = 0
g_knob_6_current_value = 0
g_knob_7_current_value = 0
g_knob_8_current_value = 0
faders_on_off = true
g_last_input_time = 0
g_last_input_item = 0
shift_status = 0
cliplaunch_1_pressed = 0
cliplaunch_2_pressed = 0
track_down_pressed = 0
track_up_pressed = 0







--------------------------------
-- update controls last value --
--------------------------------
 
function remote_on_auto_input(item_index)
   -- called after an auto-input item msg has been handled

   g_last_input_item = item_index  -- last control used
   g_last_input_time = remote.get_time_ms()
 
   if item_index == g_fader_9 then g_fader_9_last_value = remote.get_item_value(g_fader_9)
   elseif item_index == g_knob_1 then
      g_knob_1_last_value = remote.get_item_value(g_knob_1)
   elseif item_index == g_knob_2 then
      g_knob_2_last_value = remote.get_item_value(g_knob_2)
   elseif item_index == g_knob_3 then
      g_knob_3_last_value = remote.get_item_value(g_knob_3)
   elseif item_index == g_knob_4 then
      g_knob_4_last_value = remote.get_item_value(g_knob_4)
   elseif item_index == g_knob_5 then
      g_knob_5_last_value = remote.get_item_value(g_knob_5)
   elseif item_index == g_knob_6 then
      g_knob_6_last_value = remote.get_item_value(g_knob_6)
   elseif item_index == g_knob_7 then
      g_knob_7_last_value = remote.get_item_value(g_knob_7)
   elseif item_index == g_knob_8 then
      g_knob_8_last_value = remote.get_item_value(g_knob_8)
   end
end

prev_shift_status = -1
-- current status (cliplaunch and track buttons)
c_1 = 0x01
c_2 = 0x01
t_1 = 0x01
t_2 = 0x01
-- previous status (cliplaunch and track buttons)
p_c_1 = 0
p_c_2 = 0
p_t_1 = 0
p_t_2 = 0







-- to keep track of the status of the HW ctrls

local ctrl_status = {
   --       {current enabled/disabled, previous enabl/disabl, ctrl id}
   knob_1 = {false,true,"pot 0"},
   knob_2 = {false,true,"pot 1"},
   knob_3 = {false,true,"pot 2"},
   knob_4 = {false,true,"pot 3"},
   knob_5 = {false,true,"pot 4"},
   knob_6 = {false,true,"pot 5"},
   knob_7 = {false,true,"pot 6"},
   knob_8 = {false,true,"pot 7"},

   fader_1 = {false,true,"fader 0"},
   fader_2 = {false,true,"fader 1"},
   fader_3 = {false,true,"fader 2"},
   fader_4 = {false,true,"fader 3"},
   fader_5 = {false,true,"fader 4"},
   fader_6 = {false,true,"fader 5"},
   fader_7 = {false,true,"fader 6"},
   fader_8 = {false,true,"fader 7"},
   fader_9 = {false,true,"fader 8"},
}


local labels = {
   -- ctrl = {current label, previous label}
   ["pot 0"] = {"", ""},
   ["pot 1"] = {"", ""},
   ["pot 2"] = {"", ""},
   ["pot 3"] = {"", ""},
   ["pot 4"] = {"", ""},
   ["pot 5"] = {"", ""},
   ["pot 6"] = {"", ""},
   ["pot 7"] = {"", ""},
   ["fader 0"] = {"", ""},
   ["fader 1"] = {"", ""},
   ["fader 2"] = {"", ""},
   ["fader 3"] = {"", ""},
   ["fader 4"] = {"", ""},
   ["fader 5"] = {"", ""},
   ["fader 6"] = {"", ""},
   ["fader 7"] = {"", ""},
   ["fader 8"] = {"", ""},
}


local last_updated_item = false
local prev_updated_item = ""
local last_param = ""
local prev_param
local last_value = 0
local prev_value
local patch_name = false  -- when true display patch name


--------------------------
-- update machine state --
--------------------------

function remote_set_state(changed_items)

   for i, item_index in ipairs(changed_items) do

      local current_item = remote.get_item_state(item_index)
      last_param = current_item.remote_item_name
      last_value = current_item.text_value
            

      -- FADERS
      if item_index == g_fader_1 then
         g_fader_1_current_value   = remote.get_item_value(g_fader_1)
         last_updated_item = "fader 0"
         ctrl_status.fader_1[1] = current_item.is_enabled
         labels["fader 0"][1] = current_item.remote_item_name
	 
      elseif item_index == g_fader_2 then
         g_fader_2_current_value   = remote.get_item_value(g_fader_2)
         last_updated_item = "fader 1"
         ctrl_status.fader_2[1] = current_item.is_enabled
         labels["fader 1"][1] = current_item.remote_item_name
	 
      elseif item_index == g_fader_3 then
         g_fader_3_current_value   = remote.get_item_value(g_fader_3)
         last_updated_item = "fader 2"
         ctrl_status.fader_3[1] = current_item.is_enabled
         labels["fader 2"][1] = current_item.remote_item_name
	 
      elseif item_index == g_fader_4 then
         g_fader_4_current_value   = remote.get_item_value(g_fader_4)
         last_updated_item = "fader 3"
         ctrl_status.fader_4[1] = current_item.is_enabled
         labels["fader 3"][1] = current_item.remote_item_name
	 
      elseif item_index == g_fader_5 then
         g_fader_5_current_value   = remote.get_item_value(g_fader_5)
         last_updated_item = "fader 4"
         ctrl_status.fader_5[1] = current_item.is_enabled
         labels["fader 4"][1] = current_item.remote_item_name
	 
      elseif item_index == g_fader_6 then
         g_fader_6_current_value   = remote.get_item_value(g_fader_6)
         last_updated_item = "fader 5"
         ctrl_status.fader_6[1] = current_item.is_enabled
         labels["fader 5"][1] = current_item.remote_item_name
	 
      elseif item_index == g_fader_7 then
         g_fader_7_current_value   = remote.get_item_value(g_fader_7)
         last_updated_item = "fader 6"
         ctrl_status.fader_7[1] = current_item.is_enabled
         labels["fader 6"][1] = current_item.remote_item_name
	 
      elseif item_index == g_fader_8 then
         g_fader_8_current_value   = remote.get_item_value(g_fader_8)
         last_updated_item = "fader 7"
         ctrl_status.fader_8[1] = current_item.is_enabled
         labels["fader 7"][1] = current_item.remote_item_name
	 
      elseif item_index == g_fader_9 then
         g_fader_9_current_value   = remote.get_item_value(g_fader_9)
         last_updated_item = "fader 8"
         ctrl_status.fader_9[1] = current_item.is_enabled
         labels["fader 8"][1] = current_item.remote_item_name
	 


      -- KNOBS
      elseif item_index == g_knob_1 then
         g_knob_1_current_value    = remote.get_item_value(g_knob_1)
         last_updated_item = "pot 0"
         ctrl_status.knob_1[1] = current_item.is_enabled
         labels["pot 0"][1] = current_item.remote_item_name
	 
      elseif item_index == g_knob_2 then
         g_knob_2_current_value    = remote.get_item_value(g_knob_2)
         last_updated_item = "pot 1"
         ctrl_status.knob_2[1] = current_item.is_enabled
         labels["pot 1"][1] = current_item.remote_item_name
	 
      elseif item_index == g_knob_3 then
         g_knob_3_current_value    = remote.get_item_value(g_knob_3)
         last_updated_item = "pot 2"
         ctrl_status.knob_3[1] = current_item.is_enabled
         labels["pot 2"][1] = current_item.remote_item_name
	 
      elseif item_index == g_knob_4 then
         g_knob_4_current_value    = remote.get_item_value(g_knob_4)
         last_updated_item = "pot 3"
         ctrl_status.knob_4[1] = current_item.is_enabled
         labels["pot 3"][1] = current_item.remote_item_name
	
      elseif item_index == g_knob_5 then
         g_knob_5_current_value    = remote.get_item_value(g_knob_5)
         last_updated_item = "pot 4"
         ctrl_status.knob_5[1] = current_item.is_enabled
         labels["pot 4"][1] = current_item.remote_item_name
	 
      elseif item_index == g_knob_6 then
         g_knob_6_current_value    = remote.get_item_value(g_knob_6)
         last_updated_item = "pot 5"
         ctrl_status.knob_6[1] = current_item.is_enabled
         labels["pot 5"][1] = current_item.remote_item_name
	 
      elseif item_index == g_knob_7 then
         g_knob_7_current_value    = remote.get_item_value(g_knob_7)
         last_updated_item = "pot 6"
         ctrl_status.knob_7[1] = current_item.is_enabled
         labels["pot 6"][1] = current_item.remote_item_name
	 
      elseif item_index == g_knob_8 then
         g_knob_8_current_value    = remote.get_item_value(g_knob_8) 		
         last_updated_item = "pot 7"
         ctrl_status.knob_8[1] = current_item.is_enabled
         labels["pot 7"][1] = current_item.remote_item_name


      -- PADS
      elseif item_index >= g_pad_1 and item_index <= g_pad_16 then      
	      last_updated_item = "pad"
	 
      -- SCREEN
      elseif item_index == g_lcd then
         patch_name = last_value
         last_value = ""
      else
	      last_updated_item = false 
      end
   end
end





----------------------------------
-- update control surface state --
----------------------------------

function remote_deliver_midi()
   local ret_events = {}


   -- check controls NOT in use and
   -- clear their param name and value (display)
   for k, v in pairs(ctrl_status) do
      
      -- if current ~= previous and if current == false
      if v[1] ~= v[2] and v[1] == false then
         table.insert(ret_events, set_param_name("", v[3]))
         table.insert(ret_events, set_param_value("", v[3]))
      end
      
      -- prev = current
      v[2] = v[1]
   end



   -- scan screen labels looking for a change and in case update them
   for k, v in pairs(labels) do
      
      if v[1] ~= v[2] then
	     table.insert(ret_events, set_param_name(v[1], k))
	     --table.insert(ret_events, set_param_value("", k))
	     v[2] = v[1]
      end
   end
   

   -- display patch name
   if patch_name ~= false and patch_name ~= "" then
      table.insert(ret_events, set_display(0, patch_name))
      table.insert(ret_events, set_display(1, ""))
      patch_name = false
   end

   
   -- pad press - "custom display"
   if last_updated_item == "pad" then
      table.insert(ret_events, clear_display())
      table.insert(ret_events, set_display(0, last_param))
      table.insert(ret_events, set_display(1, last_value))

      prev_param = last_param
      prev_value = last_value
      
      last_updated_item = false

      return ret_events
   end

   -- if knob/fader - update ctrl value (display bottom row)
   -- when last_updated_item == false, it means it's not a fader/pad
   if last_value ~= prev_value and last_updated_item ~= false then
      table.insert(ret_events, set_param_value(last_value, last_updated_item))
      prev_value = last_value
   end
   
   last_updated_item = false


   
   return ret_events
end




-----------------------------
-- deal with incoming MIDI --
-----------------------------

function remote_process_midi(event)

   -- determine if a pad was hit
   local pad = 0
   
   
   -- if NOTE ON on channel 16
   if event[1] == 153 then
      
      -- look for specific pad ID
      if      event[2] == 40 then pad = 1
         elseif  event[2] == 41 then pad = 2
         elseif  event[2] == 42 then pad = 3
         elseif  event[2] == 43 then pad = 4
         elseif  event[2] == 48 then pad = 5
         elseif  event[2] == 49 then pad = 6
         elseif  event[2] == 50 then pad = 7
         elseif  event[2] == 51 then pad = 8
         elseif  event[2] == 36 then pad = 9
         elseif  event[2] == 37 then pad = 10
         elseif  event[2] == 38 then pad = 11
         elseif  event[2] == 39 then pad = 12
         elseif  event[2] == 44 then pad = 13
         elseif  event[2] == 45 then pad = 14
         elseif  event[2] == 46 then pad = 15
         elseif  event[2] == 47 then pad = 16
         else return false
      end
      
      
      -- if velocity is not 0
      if event[3] ~= 0 then
      	 local msg = { time_stamp = event.time_stamp, item = pad+13, value = 1 }
      	 remote.handle_input(msg)
      	 g_last_input_time = remote.get_time_ms()
      	 return true
	 
	   -- if velocity is 0
      else
   	 local msg={ time_stamp=event.time_stamp, item=pad+13, value=0 }
   	 remote.handle_input(msg)
   	 g_last_input_time=remote.get_time_ms()

	     return true
      end
      
      
  --[[     
   -- SHIFT
   elseif event[1] == 0xB6 then

      if event[2] == 0x3F then
         if event[3] ~= 0 then
            shift_status = 1
         else
            shift_status = 0
	      end
      end
      

   -- PLAY/STOP
   elseif event[1] == 0xB0 and event[2] == 0x73 then
      if shift_status == 1 then
	      local msg={ time_stamp=event.time_stamp, item=g_stop, value=event[3] }
	      remote.handle_input(msg)
      elseif shift_status == 0 then
	      local msg={ time_stamp=event.time_stamp, item=g_play, value=event[3]*127 }
	      remote.handle_input(msg) 
      end ]]
      
      -- CC on ch 16
   elseif event[1] == 191 then
      
      -- Track Left/Down button
      if event[2] == 103 and shift_status == 1 then
	 if event[3] ~= 0 then
	    track_down_pressed = 1
	 else
	    track_down_pressed = 0
	 end
	 local msg={ time_stamp=event.time_stamp, item=g_track_down, value=event[3] }
	 remote.handle_input(msg)
	 
	 -- Track Right/Up button
      elseif event[2] == 102 and shift_status == 1 then
	 if event[3] ~= 0 then
	    track_up_pressed = 1
	 else
	    track_up_pressed = 0
	 end
	 local msg={ time_stamp=event.time_stamp, item=g_track_up, value=event[3] }
	 remote.handle_input(msg)
      end
      
   end


   -- Deal with fader/pot pickup
   -- if CC on channel 1 or channel 16
   if event[1] == 176 or event[1] == 191 then
      
      -- FADERS
      if event[2] == 5 and faders_on_off then
	 if mute_control(event[3], g_fader_1_current_value, g_fader_1_last_value) then
	    local msg={ time_stamp=event.time_stamp, item=g_fader_1, value=event[3] }
	    remote.handle_input(msg)
	    g_fader_1_current_value = event[3]
	    g_fader_1_last_value = g_fader_1_current_value
	    return true
	 else
	    -- returning true, then "auto-input" is skipped
	    return true
	 end

      elseif event[2] == 6 and faders_on_off then
	 if mute_control(event[3], g_fader_2_current_value, g_fader_2_last_value) then
	    local msg={ time_stamp=event.time_stamp, item=g_fader_2, value=event[3] }
	    remote.handle_input(msg)
	    g_fader_2_current_value = event[3]
	    g_fader_2_last_value = g_fader_2_current_value
	    return true
	 else
	    -- returning true, then "auto-input" is skipped
	    return true
	 end

      elseif event[2] == 7 and faders_on_off then		
	 if mute_control(event[3], g_fader_3_current_value, g_fader_3_last_value) then
	    local msg={ time_stamp=event.time_stamp, item=g_fader_3, value=event[3] }
	    remote.handle_input(msg)
	    g_fader_3_current_value = event[3]
	    g_fader_3_last_value = g_fader_3_current_value
	    return true
	 else
	    -- returning true, then "auto-input" is skipped
	    return true
	 end

      elseif event[2] == 8 and faders_on_off then
	 if mute_control(event[3], g_fader_4_current_value, g_fader_4_last_value) then
	    local msg={ time_stamp=event.time_stamp, item=g_fader_4, value=event[3] }
	    remote.handle_input(msg)
	    g_fader_4_current_value = event[3]
	    g_fader_4_last_value = g_fader_4_current_value
	    return true
	 else
	    -- returning true, then "auto-input" is skipped
	    return true
	 end
	 
      elseif event[2] == 9 and faders_on_off then
	 if mute_control(event[3], g_fader_5_current_value, g_fader_5_last_value) then
	    local msg={ time_stamp=event.time_stamp, item=g_fader_5, value=event[3] }
	    remote.handle_input(msg)
	    g_fader_5_current_value = event[3]
	    g_fader_5_last_value = g_fader_5_current_value
	    return true
	 else
	    -- returning true, then "auto-input" is skipped
	    return true
	 end

      elseif event[2] == 10 and faders_on_off then
	 if mute_control(event[3], g_fader_6_current_value, g_fader_6_last_value) then
	    local msg={ time_stamp=event.time_stamp, item=g_fader_6, value=event[3] }
	    remote.handle_input(msg)
	    g_fader_6_current_value = event[3]
	    g_fader_6_last_value = g_fader_6_current_value
	    return true
	 else
	    -- returning true, then "auto-input" is skipped
	    return true
	 end

      elseif event[2] == 11 and faders_on_off then
	 if mute_control(event[3], g_fader_7_current_value, g_fader_7_last_value) then
	    local msg={ time_stamp=event.time_stamp, item=g_fader_7, value=event[3] }
	    remote.handle_input(msg)
	    g_fader_7_current_value = event[3]
	    g_fader_7_last_value = g_fader_7_current_value
	    return true
	 else
	    -- returning true, then "auto-input" is skipped
	    return true
	 end

      elseif event[2] == 12 and faders_on_off then
	 if mute_control(event[3], g_fader_8_current_value, g_fader_8_last_value) then
	    local msg={ time_stamp=event.time_stamp, item=g_fader_8, value=event[3] }
	    remote.handle_input(msg)
	    g_fader_8_current_value = event[3]
	    g_fader_8_last_value = g_fader_8_current_value
	    return true
	 else
	    -- returning true, then "auto-input" is skipped
	    return true
	 end

      elseif event[2] == 13 and faders_on_off then
	 if mute_control(event[3], g_fader_9_current_value, g_fader_9_last_value) then
	    local msg={ time_stamp=event.time_stamp, item=g_fader_9, value=event[3] }
	    remote.handle_input(msg)
	    g_fader_9_current_value = event[3]
	    g_fader_9_last_value = g_fader_9_current_value
	    return true
	 else
	    -- returning true, then "auto-input" is skipped
	    return true
	 end

	 
	 -- POTS
      elseif event[2] == 21 then
	    local msg={ time_stamp=event.time_stamp, item=g_knob_1, value=event[3] }
	    remote.handle_input(msg)
	    g_knob_1_current_value = event[3]
	    g_knob_1_last_value = g_knob_1_current_value
	    return true

      elseif event[2] == 22 then
	    local msg={ time_stamp=event.time_stamp, item=g_knob_2, value=event[3] }
	    remote.handle_input(msg)
	    g_knob_2_current_value = event[3]
	    g_knob_2_last_value = g_knob_2_current_value
	    return true

      elseif event[2] == 23 then
	    local msg={ time_stamp=event.time_stamp, item=g_knob_3, value=event[3] }
	    remote.handle_input(msg)
	    g_knob_3_current_value = event[3]
	    g_knob_3_last_value = g_knob_3_current_value
	    return true

      elseif event[2] == 24 then
	    local msg={ time_stamp=event.time_stamp, item=g_knob_4, value=event[3] }
	    remote.handle_input(msg)
	    g_knob_4_current_value = event[3]
	    g_knob_4_last_value = g_knob_4_current_value
	    return true

      elseif event[2] == 25 then
	    local msg={ time_stamp=event.time_stamp, item=g_knob_5, value=event[3] }
	    remote.handle_input(msg)
	    g_knob_5_current_value = event[3]
	    g_knob_5_last_value = g_knob_5_current_value
	    return true

      elseif event[2] == 26 then
	    local msg={ time_stamp=event.time_stamp, item=g_knob_6, value=event[3] }
	    remote.handle_input(msg)
	    g_knob_6_current_value = event[3]
	    g_knob_6_last_value = g_knob_6_current_value
	    return true

      elseif event[2] == 27 then
	    local msg={ time_stamp=event.time_stamp, item=g_knob_7, value=event[3] }
	    remote.handle_input(msg)
	    g_knob_7_current_value = event[3]
	    g_knob_7_last_value = g_knob_7_current_value
	    return true

      elseif event[2] == 28 then
	    local msg={ time_stamp=event.time_stamp, item=g_knob_8, value=event[3] }
	    remote.handle_input(msg)
	    g_knob_8_current_value = event[3]
	    g_knob_8_last_value = g_knob_8_current_value
	    return true
      end
      
   end


   
   return false
end
